/* Pxys - An ircu proxyscan, cachehash.c
 *
 * Bob Jenkins's modified hashtab.c to use with m_cache.c
 *
 * $Id: cachehash.c,v 1.3 2001/11/21 15:43:23 mbuna Exp $
--------------------------------------------------------------------
By Bob Jenkins.  hashtab.c

This implements a hash table.
* Keys are unique.  Adding an item fails if the key is already there.
* Keys and items are pointed at, not copied.  If you change the value
  of the key after it is inserted then hfind will not be able to find it.
* The hash table maintains a position that can be set and queried.
* The table length doubles dynamically and never shrinks.  The insert
  that causes table doubling may take a long time.
* The table length splits when the table length equals the number of items
  Comparisons usually take 7 instructions.
  Computing a hash value takes 35+6n instructions for an n-byte key.

  hcreate  - create a hash table
  hdestroy - destroy a hash table
   hcount  - The number of items in the hash table
   hkey    - key at the current position
   hkeyl   - key length at the current position
   hstuff  - stuff at the current position
  hfind    - find an item in the table
   hadd    - insert an item into the table
   hdel    - delete an item from the table
  hstat    - print statistics about the table
   hfirst  - position at the first item in the table
   hnext   - move the position to the next item in the table
--------------------------------------------------------------------
* 
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>

#ifndef STANDARD
#include "standard.h"
#endif
#ifndef LOOKUPA
#include "lookupa.h"
#endif
#ifndef PXYS_CACHEHASH_H
#include "cachehash.h"
#endif
#ifndef RECYCLE
#include "recycle.h"
#endif

#ifdef DMALLOC
#include <dmalloc.h>
#else
#include "xmalloc.h"
#endif

/*
 * cache_hgrow - Double the size of a hash table.
 * Allocate a new, 2x bigger array,
 * move everything from the old array to the new array,
 * then free the old array.
 */
static void cache_hgrow( t)
cachehtab  *t;    /* table */
{
  register ub4     newsize = (ub4)1<<(++t->logsize);
  register ub4     newmask = newsize-1;
  register ub4     i;
  register cachehitem **oldtab = t->table;
  register cachehitem **newtab = (cachehitem **)xmalloc(newsize*sizeof(cachehitem *));

  /* make sure newtab is cleared */
  for (i=0; i<newsize; ++i) newtab[i] = (cachehitem *)0;
  t->table = newtab;
  t->mask = newmask;

  /* Walk through old table putting entries in new table */
  for (i=newsize>>1; i--;)
  {
    register cachehitem *this, *that, **newplace;
    for (this = oldtab[i]; this;)
    {
      that = this;
      this = this->next;
      newplace = &newtab[(that->hval & newmask)];
      that->next = *newplace;
      *newplace = that;
    }
  }

  /* position the hash table on some existing item */
  cache_hfirst(t);

  /* free the old array */
  xfree((char *)oldtab);

}

/* cache_hcreate - create a hash table initially of size power(2,logsize) */
cachehtab *cache_hcreate(logsize)
word  logsize;    /* log base 2 of the size of the hash table */
{
  ub4 i,len;
  cachehtab *t = (cachehtab *)xmalloc(sizeof(cachehtab));

  len = ((ub4)1<<logsize);
  t->table = (cachehitem **)xmalloc(sizeof(cachehitem *)*(ub4)len);
  for (i=0; i<len; ++i) t->table[i] = (cachehitem *)0;
  t->logsize = logsize;
  t->mask = len-1;
  t->count = 0;
  t->apos = (ub4)0;
  t->ipos = (cachehitem *)0;
  t->space = remkroot(sizeof(cachehitem));
  t->bcount = 0;
  return t;
}

/* cache_hdestroy - destroy the hash table and free all its memory */
void cache_hdestroy( t)
cachehtab  *t;    /* the table */
{
  /*cachehitem *h;*/
  refree(t->space);
  xfree((char *)t->table);
  xfree((char *)t);
}

/* cache_hcount() is a macro, see cachehash.h */
/* cache_hkey() is a macro, see cachehash.h */
/* cache_hstuff() is a macro, see cachehash.h */

/* cache_hfind - find an item with a given key in a hash table */
word   cache_hfind( t, key )
cachehtab  *t;     /* table */
ub1   *key;   /* key to find */
{
  cachehitem *h;
  ub4    x = lookup(key,CACHEHASH_KEYLENGTH,0);
  ub4    y;
  for (h = t->table[y=(x&t->mask)]; h; h = h->next)
  {
    if ((x == h->hval) && 
      /*  (keyl == h->keyl) && -- CCC: same length */
        !memcmp(key, h->key, CACHEHASH_KEYLENGTH))
    {
      t->apos = y;
      t->ipos = h;
      return TRUE;
    }
  }
  return FALSE;
}

/* cache_hfindx - find an item with a given key in a hash table
 * optimized version for m_cache.c   --mbuna 15oct2001
 */
word   cache_hfindx( t, key, ohv )
cachehtab  *t;     /* table */
ub1   *key;   /* key to find */
ub4   *ohv;   /* hash value (out) */
{
  cachehitem *h;
  ub4    x = lookup(key,CACHEHASH_KEYLENGTH,0);
  ub4    y;
  *ohv = x;
  for (h = t->table[y=(x&t->mask)]; h; h = h->next)
  {
    if ((x == h->hval) && 
        /* (keyl == h->keyl) && */
        !memcmp(key, h->key, CACHEHASH_KEYLENGTH))
    {
      t->apos = y;
      t->ipos = h;
      return TRUE;
    }
  }
  return FALSE;
}

/*
 * cache_hadd - add an item to a hash table.
 * return FALSE if the key is already there, otherwise TRUE.
 */
word cache_hadd(t, key, stuff)
cachehtab  *t;      /* table */
ub1   *key;    /* key to add to hash table */
time_t stuff;  /* stuff to associate with this key */
{
  register cachehitem  *h,**hp;
  register ub4     y, x = lookup(key,CACHEHASH_KEYLENGTH,0);

  /* make sure the key is not already there */
  for (h = t->table[(y=(x&t->mask))]; h; h = h->next)
  {
    if ((x == h->hval) && 
        /* (keyl == h->keyl) && */
        !memcmp(key, h->key, CACHEHASH_KEYLENGTH))
    {
      t->apos = y;
      t->ipos = h;
      return FALSE;
    }
  }

  /* find space for a new item */
  h = (cachehitem *)renew(t->space);

  /* make the hash table bigger if it is getting full */
  if (++t->count > (ub4)1<<(t->logsize))
  {
    cache_hgrow(t);
    y = (x&t->mask);
  }

  /* add the new key to the table */
  h->key   = key;
  h->stuff = stuff;
  h->hval  = x;
  hp = &t->table[y];
  h->next = *hp;
  *hp = h;
  t->ipos = h;
  t->apos = y;
  
  return TRUE;
}

/*
 * cache_haddx - add _directly_ an item to a hash table with its known hash value.
 * return FALSE if the key is already there, otherwise TRUE.
 *
 * mbuna 15oct2001
 */
word cache_haddx( t, key, stuff, hval)
cachehtab  *t;      /* table */
ub1   *key;    /* key to add to hash table */
time_t stuff;  /* stuff to associate with this key */
ub4    hval;   /* hash value */
{
  register cachehitem  *h,**hp;
  register ub4     y;
  
  /* find space for a new item */
  h = (cachehitem *)renew(t->space);

  /* make the hash table bigger if it is getting full */
  if (++t->count > (ub4)1<<(t->logsize))
     cache_hgrow(t);
  
  y = (hval&t->mask);

  /* add the new key to the table */
  h->key   = key;
  h->stuff = stuff;
  h->hval  = hval;
  hp = &t->table[y];
  h->next = *hp;
  *hp = h;
  t->ipos = h;
  t->apos = y;
  
  return TRUE;
}


/* cache_hdel - delete the item at the current position */
word cache_hdel(t)
cachehtab *t;      /* the hash table */
{
  cachehitem  *h;    /* item being deleted */
  cachehitem **ip;   /* a counter */

  /* check for item not existing */
  if (!(h = t->ipos)) return FALSE;

  /* remove item from its list */
  for (ip = &t->table[t->apos]; *ip != h; ip = &(*ip)->next)
    ;
  *ip = (*ip)->next;
  --(t->count);

  /* adjust position to something that exists */
  if (!(t->ipos = h->next)) cache_hnbucket(t);

  /* recycle the deleted cachehitem node */
  redel(t->space, h);
  
  return TRUE;
}

/* cache_hfirst - position on the first element in the table */
word cache_hfirst(t)
cachehtab  *t;    /* the hash table */
{
  t->apos = t->mask;
  (void)cache_hnbucket(t);
  return (t->ipos != (cachehitem *)0);
}

/* hnext() is a macro, see hashtab.h */

/*
 * cache_hnbucket - Move position to the first item in the next bucket.
 * Return TRUE if we did not wrap around to the beginning of the table
 */
word cache_hnbucket(t)
cachehtab *t;
{
  ub4  oldapos = t->apos;
  ub4  end = (ub4)1<<(t->logsize);
  ub4  i;

  /* see if the element can be found without wrapping around */
  for (i=oldapos+1; i<end; ++i)
  {
    if (t->table[i&t->mask])
    {
      t->apos = i;
      t->ipos = t->table[i];
      return TRUE;
    }
  }

  /* must have to wrap around to find the last element */
  for (i=0; i<=oldapos; ++i)
  {
    if (t->table[i])
    {
      t->apos = i;
      t->ipos = t->table[i];
      return FALSE;
    }
  }

  return FALSE;
}

void cache_hstat(t, obuckets, oitems)
cachehtab *t;
ub4* obuckets;
ub4 *oitems;
{
  /* store statistics */
  *obuckets = ((ub4)1<<t->logsize);
  *oitems = t->count;
}
