/* Pxys - An IRC ProxyScan, m_array.c
 * 
 * Copyright (C) 2001 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 *
 * Threaded IRC remote clients array functions.
 *
 * $Id: m_array.c,v 1.32 2001/11/21 15:43:23 mbuna Exp $
 */


#include "config.h"

#include <sys/time.h> // tmp

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "m_array.h"
#include "m_clientq.h"
#include "m_structs.h"
#include "m_log.h"
#include "m_cache.h"
#include "m_glineq.h"
#include "m_noscan.h"
#include "m_proxycache.h"
#include "m_share.h"
#include "m_thread.h"

#include "ms_parser.h"
#include "ms_globals.h"
#include "ms_msgout.h" /* 1.6 */
#include "ms_netutils.h"
#include "ms_network.h" /* 1.6 */

#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "recycle.h"
#include "misc.h"
#include "numnicks.h"
#include "strnlwrcpy.h"
#include "typedefs.h"
#include "yxx.h"
/* xmalloc at the bottom of include files */
#include "xmalloc.h"


/* Global variables. */

long clients_counter = 0; /* number of call of array_addclient */

long ipa_current = 0;              /* Current size of the array */
long ipa_max = 0;                  /* Maximum size of the array */



struct Client *mClient = NULL;
struct Client *mClientQ = NULL;
struct Client *mLastClient = NULL;

htab *mHC = NULL;
extern htab *mHS;
extern server_handle_t mSA;
extern scanstats_s *scanstats;
extern time_t startscan_time;

/* Using mutex to keep the array sorted without conflict with the threads.
 * mutex_link is the main mutex which protects the array critical section (when
 * the array linking is modified). Threads which want to check the
 * array need to wait until these zone are unlocked.
 */
pthread_mutex_t mutex_link;
pthread_mutex_t mutex_shutdown;
pthread_mutex_t mutex_scanext;
extern pthread_mutex_t mutex_server;

extern scanthread_s *scanthread;

/* _init the array
 * After the mutex initialized, do not forget to allocate the outflow structs feature.
 */
void
array_init()
{
   /* Array mutex lockers */
   pthread_mutex_init(&mutex_link, NULL);
   pthread_mutex_init(&mutex_shutdown, NULL);
   pthread_mutex_init(&mutex_scanext, NULL);
   
   /* Create main clients hash table */
   mHC = jenkins_hcreate(2);
}

void
array_destroy()
{
   if (mHC) {
      jenkins_hdestroy(mHC);
      mHC = NULL;
      pthread_mutex_destroy(&mutex_link);
      pthread_mutex_destroy(&mutex_shutdown);
      pthread_mutex_destroy(&mutex_scanext);
   }
}

/* array_addclient is always called by the same (main) thread (the one the IRC socket is).
 */
void
array_addclient(nick, mode, b64IP, numnick, ident)
const char *nick;       /* Client nickname */
const char *mode;       /* Client mode or NULL */
const char *b64IP;      /* Client base64 IP */
const char *numnick;    /* Client numeric */
int ident;              /* 1 if idented, 0 if not (~user) */
{
   yxx target_yxx;
   int cached = 0;
   proxycacheItem *citem;
   struct Server *sptr = NULL;
   struct Client *cptr = (struct Client *) xmalloc(sizeof(struct Client));
   
   /* Fill Client struct... */
   target_yxx = yxx_to_int(numnick);
   cptr->nserv = target_yxx.server_n;
   cptr->nnick = target_yxx.client_n;
   strnlwrcpy(cptr->nick, nick, NICKLEN);
   cptr->nick[NICKLEN] = '\0';
   
   cptr->flags = FLAG_NULL;
   if (ident)
      cptr->flags |= FLAG_IDENT;
   if (mode && strchr(mode, 'o'))
      cptr->flags |= FLAG_OPER;
   
   cptr->ip.s_addr = htonl(base64toint(b64IP));
   cptr->last = 0;
   
   pthread_mutex_lock(&mutex_link);
   
   /* Find server; get clients array; link. */
   if (target_yxx.server_n < NN_MAX_SERVER)
      sptr = mSA[target_yxx.server_n];
   if(NULL == sptr) {
      log_error("array_addclient: server not found (%d)\n", target_yxx.server_n);
      pthread_mutex_unlock(&mutex_link);
      return;
   }
   
   sptr->cliH[target_yxx.client_n & sptr->nn_mask] = (void*)cptr;
   sptr->clients++;
   
   hadd(mHC, cptr->nick, strlen(cptr->nick) + 1, (void*)cptr);
   
   ++ipa_current;
   if(ipa_max < ipa_current)
      ipa_max = ipa_current;   /* Only here for fun */
   
   /* Another thingy... proxycache, new in 1.5
    * Rebuilt in 1.6 to support glineq...
    */
   
   if (proxycache_check(&cptr->ip, &citem)) {
      if (is_bursting(sptr)) {
         char *typeproxy, *reason;
         get_proxy_type(citem->type, 1, &typeproxy, &reason);
         if (reason && glineq_push(&citem->addr, citem->type))
            client_sendmsgto_console("%s *@%s [1+] %s (cached+queued)", PROXY_GLINE_CMD,
               inet_ntoa(citem->addr), typeproxy);
      } else {
         /* Warning: avoid dead lock here! */
         pthread_mutex_unlock(&mutex_link);      /* UNLOCK */
         proxy_toast(&cptr->ip, citem->type, 1); /* proxy_toast might lock mutex_link! */
         pthread_mutex_lock(&mutex_link);        /* LOCK */
      }
   }
   
   ++clients_counter;
   
   /* ClientQ support: recent connected users */
   if (!cached && scanstats->is_sync)
      clientq_queue(cptr->nick);
   
   pthread_mutex_unlock(&mutex_link);
}


/* array_delclient is always called by the same (main) thread.
 * Used to delete only 1 client (after a QUIT or KILL message).
 */
void
array_delclient(const char *numnick)
{
   struct Server *sptr;
   struct Client *cptr;
   yxx target_yxx;
   
   target_yxx = yxx_to_int(numnick);
   
   Debug((DBLMAX, "array_delclient> %d %d", target_yxx.server_n, target_yxx.client_n));
   
   pthread_mutex_lock(&mutex_link);
   
   sptr = mSA[target_yxx.server_n];
   if(NULL == sptr) {
      log_error("array_delclient: server not found (%d)", target_yxx.server_n);
      return;
   }
   
   cptr = sptr->cliH[target_yxx.client_n & sptr->nn_mask];
   if(NULL == sptr) {
      log_error("array_delclient: client not found (%d %d)",
           target_yxx.server_n, target_yxx.client_n);
      xfree(cptr);
      return;
   }
   
   if(hfind(mHC, cptr->nick, strlen(cptr->nick) + 1)) {
      hdel(mHC);
   } else {
      pthread_mutex_unlock(&mutex_link);
      log_error("array_delclient: hashed client not found (%d %d) nick: %s (num: %s)",
           target_yxx.server_n, (target_yxx.client_n & sptr->nn_mask), cptr->nick, numnick);
      xfree(cptr);
      return;
   }
   
   sptr->cliH[target_yxx.client_n & sptr->nn_mask] = NULL;
   sptr->clients--;
   --ipa_current;
   
   pthread_mutex_unlock(&mutex_link);
   
   Debug((DBLMAX, "array_delclient> Successfull (%s)", cptr->nick));
   
   xfree(cptr);
}


/* array_purgeserv is always called by the same (main) thread.
 * Used to free all users from a specified server numeric.
 *
 * I rewrote a low level dealing of the hash table; that's why:
 *
 * "The most unusual thing about this module (hashtab.c) is that it maintains
 *  a current position. This means you can't have a dangling pointer into
 *  it. If you position on something, then delete it, the position
 *  automatically moves on to some other item. On the other hand, it
 *  also means it's hard to do nested loops over all the items in the
 *  table, because there is only one position at a time. " -- Bob Jenkins
 * 
 *
 * --mbuna 20010327
 */

void
array_purgeserv(unsigned char yy)
{
   struct Client *cptr;
   struct Server *sptr;
   ub4 i, end, counter;
   hitem  *h, *hnext;
   hitem **ip;   /* a counter */
   
   Debug((DBLMAX, "array_purgeserv> yy == %d", (int)yy));
   
   pthread_mutex_lock(&mutex_link);
   
   sptr = mSA[yy];
   
   end = (ub4)1<<(mHC->logsize);
   counter = 0;
   for (i = 0; i < end;  ++i) {
      for (h=mHC->table[i]; h; h=hnext) {
         hnext = h->next;
         
         cptr = (struct Client *)h->stuff;
         
         if(cptr->nserv == yy) {
            
            /* hash table: remove item from its list */
            for (ip = &mHC->table[i]; *ip != h; ip = &(*ip)->next)
               ;
            *ip = (*ip)->next;
            --(mHC->count);
            
             /* hash table: recycle the deleted hitem node */
            redel(mHC->space, h);
            
            sptr->cliH[cptr->nnick & sptr->nn_mask] = NULL;
            sptr->clients--;
            --ipa_current;
            
            xfree(cptr);
         }
      }
   }
   
   /* hsanity(mHC); */
   
   /* warning: ipos not in apos -> adjust position */
   (void) hnbucket(mHC);
   
   pthread_mutex_unlock(&mutex_link);
   
   Debug((DBLMAX, "array_purgeserv> Successfull (%d) - hcount: %ld", (int)yy, hcount(mHC)));
}


/* Called from the main thread only. */
void
array_nickchange(const char *numnick, const char *newnick)
{
   struct Server * sptr;
   struct Client * cptr;
   yxx target_yxx;
   
   target_yxx = yxx_to_int(numnick);
   
   sptr = mSA[target_yxx.server_n];
   if (NULL == sptr) {
      log_error("array_nickchange: Server #%d not found!", target_yxx.server_n);
      return;
   }
   
   /* Server found. */
   cptr = sptr->cliH[target_yxx.client_n & sptr->nn_mask];
   if (NULL == cptr) {
      log_error("array_nickchange: Client ID %s (%d) not found!",
           numnick, target_yxx.client_n & sptr->nn_mask);
      return;
   }
   
   /* Same nick ? */
   if (stricmp(cptr->nick, newnick) == 0)
      return;
   
   /* Client found. */
   pthread_mutex_lock(&mutex_link);
   
   if (hfind(mHC, cptr->nick, strlen(cptr->nick) + 1)) {
      /* Remove it from hash table, as we hash on nicks. */
      hdel(mHC);
      /* Update Client struct. Only deal with lowercase nicks. */
      strnlwrcpy(cptr->nick, newnick, NICKLEN);
      cptr->nick[NICKLEN] = '\0';
      
      /* Add it to hash table. */
      hadd(mHC, cptr->nick, strlen(newnick) + 1, (void*)cptr);
   } else
      log_error("array_nickchange: hfind failed, nick %s not found on %s",
           cptr->nick, sptr->name);
   
   pthread_mutex_unlock(&mutex_link);
}


/* array_fetchitem is called by EVERY scan threads!
 * Used to get/set the next targets for the thread th.
 * Pre-condition: th != NULL.
 */
void
array_fetchitem(scanthread_s * th)
{
   struct Server *serv = NULL;
   struct Client *cli = NULL;
   int i, lookup, result, except;
   unsigned int npos = 0; /* shut up gcc */
   
   th->counter = 0;
   
   
   /* Cleaning */
   if (th->state & STH_STATE_RUNNING)
      for (i = 0; i < SCANS_PER_THREAD; i++)
         th->item[i].state = SIT_INACTIVE;
            
   pthread_mutex_lock(&mutex_link); /* suspend any clients list
                                     * or clientq changes during the fetch
                                     */
   
   if (scanstats->is_sync) {
      /* We are sync (not working at 100% or our threads capacity,
       * so, we only check for clients in clientq (new clients).
       * That's faster than scanning the whole network's users, but,
       * remember it's only possible when the pxys is sync.
       */
      for(i = 0; i < SCANS_PER_THREAD; i++) {
         
         result = array_fetchitem_clientq(&cli);
         
         if (result) {
            Debug((DBLMAX, "array_fetchitem> is_sync array_fetchitem_clientq OK"));
            th->item[i].nserv = cli->nserv;
            th->item[i].nnick = cli->nnick;
            memcpy(&th->item[i].addr, &cli->ip, sizeof(struct in_addr));
            th->item[i].state = SIT_ACTIVE;
            th->counter++;
         } else
            /* no more clients in clientq, this is the end */
            break;
      }
      if (th->counter == 0) {
         /* We didn't found any clients in clientq, and we are sync? It simply
          * means no new client has connected. Thus, this thread is going to wait now.
          */
         if (th->state & STH_STATE_RUNNING) {
            /* Change scanthread state */
            scanstats->running_threads--;
            scanstats->waiting_threads++;
            th->state = STH_STATE_WAITING;
            Debug((DBLMAX, "array_fetchitem> is_sync STH_STATE_WAITING"));
         }
         /* else.. This thread was already "waiting" for something to scan */
      } else {
         /* We found at least 1 IP to scan */
         if (th->state & STH_STATE_WAITING) {
            scanstats->running_threads++;
            scanstats->waiting_threads--;
         }
         th->state = STH_STATE_RUNNING;
      }
   }
   else { /* scanstats->is_sync */
      
      if (th->state & STH_STATE_RUNNING) {
         
         /* First check on the "current" thread's server if there are users remaining,
          * then, if not found, we try the next server(s), until we found something.
          * If really nothing is found, the pxys's just sync! That's cool, we'll be able
          * to use the clientq now, which will save CPU here.
          */
         
         serv = mSA[th->item[SCANS_PER_THREAD - 1].nserv]; /* Server Array */
         npos = th->item[SCANS_PER_THREAD - 1].nnick;
         
         
      } else
         serv = NULL;
      
         
      /* Server still there? If NULL, it's in split (or thread in waiting state...) */
      if (!serv) {
         hfirst(mHS);
         serv = (struct Server *)hstuff(mHS);
         npos = 0;
      }
      
      lookup = 0;
      for (i = 0; i < SCANS_PER_THREAD; i++) {   /* for each IP to scan in a scanning thread */
         for ( ; ; ) {                           /* do...while we have not lookup'd all servers */
            
                                                 /* Check for server scan exceptions */
            check_noscan_server(serv->myx, &except);
            
            if (!except)
               check_shared_server(serv->myx, &except);
           
                                                 /* Lookup the server's users for IPs not scanned */
            if (!except)
               result = array_fetchitem_scanserv(serv, npos, &cli);

                                                 /* User found, go to next IPs in the scanthread */
            if (!except && result) {
               th->item[i].nserv = cli->nserv;
               th->item[i].nnick = cli->nnick;
               memcpy(&th->item[i].addr, &cli->ip, sizeof(struct in_addr));
               th->item[i].state = SIT_ACTIVE;
               th->counter++;
               break;
               
            } else {
                                                 /* User not found, try next server in the table */
               if (hfind(mHS, serv->name, strlen(serv->name) + 1) && hnext(mHS))
                  serv = (struct Server *)hstuff(mHS);
               
               else if(hfirst(mHS) && (hstuff(mHS) != (void*)serv))
                  serv = (struct Server *)hstuff(mHS);
               
               else
                                                 /* Looks like the pxys is no longer linked */
                  goto exitfor;
                                                 /* Check client from 0 to serv->nn_mask */
               npos = 0;
               
                                                 /* We are sure we scanned the whole network,
                                                  * so why continuing trying to fill other IPs in thread..
                                                  */
               if (lookup++ >= hcount(mHS))
                  goto exitfor;
            }
         }
      } /* for (i = 0; i < SCANS_PER_THREAD; i++) */
exitfor:
      if (th->counter == 0) {
         /* We're sync! */
         scanstats->is_sync = 1;
         
         /* Change scanthread state */
         if (th->state & STH_STATE_RUNNING) {
            scanstats->running_threads--;
            scanstats->waiting_threads++;
         }
         th->state = STH_STATE_WAITING;
         
      } else {
         if (th->counter < SCANS_PER_THREAD)
            /* We're sync! But there are still some users to scan... */
            scanstats->is_sync = 1;
         
         if (th->state & STH_STATE_WAITING) {
            scanstats->waiting_threads--;
            scanstats->running_threads++;
         }
         th->state = STH_STATE_RUNNING;
      }
   } /* scanstats->is_sync */
   pthread_mutex_unlock(&mutex_link);
}

/* RETURNS
 * 1 if a valid client to scan is found
 * 0 if no client to scan remains in clientq
 */
int
array_fetchitem_clientq(struct Client **cliH)
{
   char *nick;
   int noscan;
   struct Client *cli = NULL;
   time_t now, cached;
   
   time(&now);
   
   for (;;) {
      /* deque until a client is still online (or no more clients are in queue) */
      if (!clientq_deque(&nick))
         break;
      cli = array_findclient_from_nick(nick);
      
      if (cli) {
         check_noscan_server(cli->nserv, &noscan);
         if (!noscan)
            check_shared_server(cli->nserv, &noscan);
         
         if (noscan)
            continue;
         
         if (!cli->last) {
            /* client not checked */
            
            cli->last = now;
            
            /* "protected" by noscan ? */
            check_noscan_ip(&cli->ip, &noscan);
            
            if (noscan)
               continue;
            
            /* Check cache */
            cached = cache_fetch(&cli->ip);
            
            if (!cached || (now - cached > MyPerform->recheck)) {
               cache_insert(&cli->ip, now);
               *cliH = cli;
               return 1;
            }
         }
      }
   }
   /* Boom, nothing valid found */
   return 0;
}


int
array_fetchitem_scanserv(
   struct Server *serv,
   unsigned int last_client_pos,
   struct Client **cliH)
{
   unsigned int u;
   int result, noscan;
   time_t cached;
   time_t now;
   struct Client *cli = NULL;
   
   pthread_mutex_lock(&mutex_server); /* Do NOT allow any servers change */
   
   time(&now);
   result = 0;
   
   for (u = last_client_pos; u <= serv->nn_mask; u++) {
      if((cli = (struct Client *)serv->cliH[u & serv->nn_mask])) {
         /* We found a client using this slot, scan him?
          */
         if (!cli->last) {
            /* client not checked */
            
            cli->last = now;
            
            /* Hrms, "protected" by noscan ? */
            check_noscan_ip(&cli->ip, &noscan);
            
            if (noscan)
               continue;
            
            /* Check cache */
            cached = cache_fetch(&cli->ip);
            if(!cached || ((now - cached) > MyPerform->recheck)) {
               cache_insert(&cli->ip, now);
               *cliH = cli;
               result = 1;
               break;
            }
         }
      }
   }
   pthread_mutex_unlock(&mutex_server);
   return result;
}
            


/* Might be called by the other threads */
long
array_countclient_from_ip(struct in_addr *addr)
{
	long counter = 0;
    struct Client *cptr;
    
	pthread_mutex_lock(&mutex_link);
	
	if (hfirst(mHC)) do
	{
	   cptr = (struct Client *)hstuff(mHC);
       if (addr->s_addr == cptr->ip.s_addr) /*memcmp(addr, &cptr->ip, sizeof(struct in_addr)) == 0)*/
          ++counter ;
       
    } while(hnext(mHC));
    
	pthread_mutex_unlock(&mutex_link);
	
	return counter;
}

/* Have to be called from the 'main thread', the same thread which is calling
 * addclient and delclient (the IRC socket thread). */
struct Client *
array_findclient_from_numeric(const char *numnick)
{
   struct Server *sptr;
   yxx client_yxx;
   
   client_yxx = yxx_to_int(numnick);
   sptr = mSA[client_yxx.server_n];
   if (NULL == sptr)
      return NULL;
   
   return sptr->cliH[client_yxx.client_n & sptr->nn_mask];
}

/*
struct Client *
array_findclient_from_ip(struct in_addr *iptr)
{
   struct Client *cptr = NULL;
   
   if(NULL == iptr)
      return NULL;
     
   cptr = mClient;
   for(; cptr; cptr = cptr->next) {
      if(memcmp(&cptr->ip, iptr, sizeof(struct in_addr)) == 0)
         break;
   }
   
   return cptr;
}
*/

/* Called from the main thread only. */
struct Client *
array_findclient_from_nick(const char *nick)
{
   char tmp_nick[NICKLEN+1];
   
   strnlwrcpy(tmp_nick, nick, NICKLEN);
   tmp_nick[NICKLEN] = '\0';
   
   if (hfind(mHC, tmp_nick, strlen(nick) + 1))
      return (struct Client *)hstuff(mHC);
   
   return NULL;
}




/* m_array.c */
