/* Pxys - An IRC ProxyScan, mp_listener.c
 * 
 * Copyright (C) 2001 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *
 * Target server, working along with mp_proxy.
 * Module checking proxy/webserver tunelling vulnerability.
 *
 * $Id: mp_listener.c,v 1.13 2001/11/23 13:53:12 mbuna Exp $
 */



#define PORTRANGE_MIN       2048
#define PORTRANGE_MAX       3072

#define STARTPORT           2048

#include "config.h"

#include <stdio.h>
#include <stdlib.h>   /* rand */
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <errno.h>
#include <signal.h>


#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "typedefs.h"
#include "conf.h"
#include "debug.h"

#include "m_log.h"
#include "m_thread.h"
#include "m_shutdown.h"

#include "mp_listener.h"
#include "mp_proxy.h"
#include "mp_utils.h"

#include "ms_globals.h"
#include "ms_msgout.h"

/* Globals */

int SPort;
int ls;
int startdisconnect;

void
listener_init()
{
   SPort = STARTPORT;
}


void
listener_bind(struct sockaddr_in * servAddr)
{
   int retval, err, maxtry = 0;
   do {
      
      Debug((DBLMAX, "listener_bind: port %d", SPort));
      
      servAddr->sin_port = htons(SPort);
      
      retval = bind(ls, (struct sockaddr *) servAddr, sizeof(struct sockaddr_in));
      ++maxtry;
      
      if(retval < 0) {
         err = errno;
         
         Debug((DBLMAX, "listener_bind: Cannot bind to %d (errno == %d (%s))",
            SPort, err, strerror(err)));
         
         fprintf(stderr, "listener_bind: bind failed on port %d (%s)\n",
            SPort, strerror(err));
         
         log_error("listener_bind: bind failed on port %d (%s)", SPort, strerror(err));
         
         if(maxtry > 10) {
            /* We tried 10 times and its still not working */
            myclose(ls);
            log_error("listener_bind: bind failed - aborting");
            exit(1);
         }
         
         if (err == EADDRINUSE) {
            listener_switchport(); /* Try another port... */
            continue;
         }
         
         myclose(ls);
         exit(1);
      }
      
   } while (retval < 0);
}

void
listener_run()
{
   int as, port;
   struct sockaddr_in servAddr;
   struct sockaddr_in sinOut;
   unsigned int addrlen;                 /* socklen_t */
   unsigned int opt;
   scanthread_s *th;
   unsigned int len = sizeof(struct sockaddr_in);

   /* Create socket */
   ls = socket(AF_INET, SOCK_STREAM, 0);
   if(ls < 0) {
      fprintf(stderr, "mp_listener:listener_run: socket failed.");
      log_error("mp_listener:listener_run: socket failed.\n");
      exit(EXIT_FAILURE);
   }
   
   memset(&servAddr, 0, sizeof(struct sockaddr_in));
   servAddr.sin_family = AF_INET;
   servAddr.sin_addr.s_addr = Me->listeneraddr;
   servAddr.sin_port = htons(SPort);
   
   /* Allow fast rebind */
   opt = 1;
   setsockopt(ls, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
   
   /* Binding */
   listener_bind(&servAddr);
   
   /* listen for connections, backlog should be able to handle "massive" replies */
   listen(ls, 16);
   
   printf("Listener running at %s:%d (http proxy check).\n", Me->listener, SPort);
   
   /* Detach the listener thread, so that we can continue scanner_init */
   pthread_detach(pthread_self());
   
   /**DETACHED**/
   
   startdisconnect = 0;
   
   while (!startdisconnect) {
      
      Debug((DBLMAX, "mp_listener:listener_run: accept()"));
      
      addrlen = sizeof(servAddr);
      as = accept(ls, (struct sockaddr *) &servAddr, &addrlen);
      
      if (as < 0) {
         /* accept() failed. */
         if (ECONNABORTED == errno) {
            /* A connection arrived, but it was closed while waiting
             * on the listen queue.
             */
            log_slacken(LOGSLACK_ACCEPT_ECONNABORTED,
                        "listener_run: accept() failed errno=%d (%s)", errno, strerror(errno));
            continue;
         }
         if (EINTR == errno)
            continue;
         log_error("mp_listener:listener_run: accept() returned error %d (%s)",
            errno, strerror(errno));
         break;
         
      } else {
         
         Debug((DBLMAX, "mp_listener:listener_run: Accepted a connection"));
         
         if (!getpeername(as, (struct sockaddr *) &sinOut, &len)) {
            /* success */

            Debug((DBLMAX, "mp_listener:listener_run: %s Proxy Anon and vulnerable to tunneling",
                   inet_ntoa(servAddr.sin_addr)));

            /* Which thread is currently handling this client? */
            
            th = get_scanthread_from_ip(&servAddr.sin_addr, &port);
            
            if (th) {
               switch (port) {
                  case 3128:
                     proxy_toast(&sinOut.sin_addr, PROXY_TYPE_HTTPPROXY_3128, 0);
                     break;
                  case 8080:
                     proxy_toast(&sinOut.sin_addr, PROXY_TYPE_HTTPPROXY_8080, 0);
                     break;
                  case 80:
                     proxy_toast(&sinOut.sin_addr, PROXY_TYPE_HTTPPROXY_80, 0);
                     break;
               }
               
               if (th->state & SCANTHREAD_TIMEWAIT)
                  pthread_cond_signal(&th->sem);   /*  Wake up thread _once_ */

            } else {
               /* Incomming connection but no associated thread found?
                * 1- The thread which was handling this client has timeout
                * 2- Someone is doing something nasty.
                */
               
               Debug((DBLMAX, "listener_run: Thread NOT found!!!"));
               
               log_slacken(LOGSLACK_SCANTHREAD_NOTFOUND,
                           "listener_run: SCANTHREAD NOT FOUND - IP: %s",
                           inet_ntoa(sinOut.sin_addr));
            }

         } else {
            myclose(as);
            
            if ((ECONNRESET == errno) || (ENOTCONN == errno)) {
               /* The connection has been reset by the p33r. */
               log_slacken(LOGSLACK_GETPEERNAME_ECONNRESET,
                           "listener_run: getpeername failed errno=%d (%s)", errno, strerror(errno));
               continue;
            }
            /* Fatal */
            log_error("mp_listener:listener_run: getpeername failed with error %d - aborting (%s)",
               errno, strerror(errno));
            break;
         }
         myclose(as);
      }
   }
   myclose(ls);
   m_shutdown();
}

void
listener_stop()
{
   Debug((DBLMAX, "Stopping listener."));
   startdisconnect = 1;
   myclose(ls);
}

void
listener_switchport()
{
   int x;
   
   Debug((DBLMAX, "Switching listener port."));
   
   /* srand already called in main()
    * **srand((unsigned int) time(NULL));**
    */
   
   do {
      x = PORTRANGE_MIN;
      x += (int) ((rand() * (double) (PORTRANGE_MAX - PORTRANGE_MIN))
      	/ ((double) (RAND_MAX + 1.)));
   } while(x == SPort);
   
   SPort = x;
   
   Debug((DBLMAX, "New listener port is %d", x));
}

void
listener_isbeingflooded()
{
   /* We need a fast port switch, someone is flooding on the current listener port.
    * We can't wait for current scan timeout, that's too risked =(
    * But... we dont like to miss someone ;)
    * Set all current scan as http proxy unscanned...
    */
   
   /* First thing to do, close the port. */
   listener_stop();
   
   /* Then, mark all running threads as 'rescan', so they will be able
    * to rescan these clients with our new good open port.
    */
   set_globalrescan();
   
   
   
   /* TODO... */
   
   
}

/* mp_listener.c */
